from collections.abc import Mapping
from typing import TYPE_CHECKING, AbstractSet, Any, Optional, Union  # noqa: UP035

import dagster._check as check
from dagster._annotations import public
from dagster._core.definitions.composition import PendingNodeInvocation
from dagster._core.definitions.decorators.graph_decorator import graph
from dagster._core.definitions.dependency import Node
from dagster._core.definitions.hook_definition import HookDefinition
from dagster._core.definitions.op_definition import OpDefinition
from dagster._core.definitions.resource_definition import IContainsGenerator, Resources
from dagster._core.errors import DagsterInvalidPropertyError, DagsterInvariantViolationError
from dagster._core.execution.context.system import StepExecutionContext
from dagster._core.execution.plan.step import ExecutionStep
from dagster._core.execution.plan.utils import RetryRequestedFromPolicy
from dagster._core.log_manager import DagsterLogManager

if TYPE_CHECKING:
    from dagster._core.instance import DagsterInstance


def _property_msg(prop_name: str, method_name: str) -> str:
    return (
        f"The {prop_name} {method_name} is not set when a `HookContext` is constructed from "
        "`build_hook_context`."
    )


def _check_property_on_test_context(
    context: "HookContext", attr_str: str, user_facing_name: str, param_on_builder: str
):
    """Check if attribute is not None on context. If none, error, and point user in direction of
    how to specify the parameter on the context object.
    """
    value = getattr(context, attr_str)
    if value is None:
        raise DagsterInvalidPropertyError(
            f"Attribute '{user_facing_name}' was not provided when "
            f"constructing context. Provide a value for the '{param_on_builder}' parameter on "
            "'build_hook_context'. To learn more, check out the testing hooks section of Dagster's "
            "concepts docs: https://legacy-docs.dagster.io/concepts/ops-jobs-graphs/op-hooks#testing-hooks"
        )
    else:
        return value


class HookContext:
    """The ``context`` object available to a hook function on an DagsterEvent."""

    def __init__(
        self,
        step_execution_context: StepExecutionContext,
        hook_def: HookDefinition,
    ):
        self._step_execution_context = step_execution_context
        self._hook_def = check.inst_param(hook_def, "hook_def", HookDefinition)
        self._required_resource_keys = hook_def.required_resource_keys
        self._resources = step_execution_context.scoped_resources_builder.build(
            self._required_resource_keys
        )

    @public
    @property
    def job_name(self) -> str:
        """The name of the job where this hook is being triggered."""
        return self._step_execution_context.job_name

    @public
    @property
    def run_id(self) -> str:
        """The id of the run where this hook is being triggered."""
        return self._step_execution_context.run_id

    @public
    @property
    def hook_def(self) -> HookDefinition:
        """The hook that the context object belongs to."""
        return self._hook_def

    @public
    @property
    def instance(self) -> "DagsterInstance":
        """The instance configured to run the current job."""
        return self._step_execution_context.instance

    @property
    def op(self) -> Node:
        """The op instance associated with the hook."""
        return self._step_execution_context.op

    @public
    @property
    def step_key(self) -> str:
        """The key for the step where this hook is being triggered."""
        return self._step_execution_context.step.key

    @public
    @property
    def required_resource_keys(self) -> AbstractSet[str]:
        """Resources required by this hook."""
        return self._required_resource_keys

    @public
    @property
    def resources(self) -> "Resources":
        """Resources available in the hook context."""
        return self._resources

    @public
    @property
    def op_config(self) -> Any:
        """The parsed config specific to this op."""
        op_config = self._step_execution_context.resolved_run_config.ops.get(
            str(self._step_execution_context.step.node_handle)
        )
        return op_config.config if op_config else None

    # Because of the fact that we directly use the log manager of the step, if a user calls
    # hook_context.log.with_tags, then they will end up mutating the step's logging tags as well.
    # This is not problematic because the hook only runs after the step has been completed.
    @public
    @property
    def log(self) -> DagsterLogManager:
        """Centralized log dispatch from user code."""
        return self._step_execution_context.log

    @public
    @property
    def op_exception(self) -> Optional[BaseException]:
        """The thrown exception in a failed op."""
        exc = self._step_execution_context.step_exception

        if isinstance(exc, RetryRequestedFromPolicy):
            return exc.__cause__

        return exc

    @public
    @property
    def op_output_values(self) -> Mapping[str, Union[Any, Mapping[str, Any]]]:
        """The computed output values.

        Returns a dictionary where keys are output names and the values are:
            * the output values in the normal case
            * a dictionary from mapping key to corresponding value in the mapped case
        """
        results: dict[str, Union[Any, dict[str, Any]]] = {}
        captured = self._step_execution_context.step_output_capture

        if captured is None:
            check.failed("Outputs were unexpectedly not captured for hook")

        # make the returned values more user-friendly
        for step_output_handle, value in captured.items():
            if step_output_handle.mapping_key:
                if results.get(step_output_handle.output_name) is None:
                    results[step_output_handle.output_name] = {
                        step_output_handle.mapping_key: value
                    }
                else:
                    results[step_output_handle.output_name][step_output_handle.mapping_key] = value
            else:
                results[step_output_handle.output_name] = value

        return results

    @public
    @property
    def op_output_metadata(self) -> Mapping[str, Union[Any, Mapping[str, Any]]]:
        """The applied output metadata.

        Returns a dictionary where keys are output names and the values are:
            * the applied output metadata in the normal case
            * a dictionary from mapping key to corresponding metadata in the mapped case
        """
        results: dict[str, Union[Any, dict[str, Any]]] = {}
        captured = self._step_execution_context.step_output_metadata_capture

        if captured is None:
            check.failed("Outputs were unexpectedly not captured for hook")

        # make the returned values more user-friendly
        for step_output_handle, metadata in captured.items():
            if step_output_handle.mapping_key:
                if results.get(step_output_handle.output_name) is None:
                    results[step_output_handle.output_name] = {
                        step_output_handle.mapping_key: metadata
                    }
                else:
                    results[step_output_handle.output_name][step_output_handle.mapping_key] = (
                        metadata
                    )
            else:
                results[step_output_handle.output_name] = metadata

        return results


class UnboundHookContext(HookContext):
    def __init__(
        self,
        resources: Mapping[str, Any],
        op: Optional[Union[OpDefinition, PendingNodeInvocation]],
        run_id: Optional[str],
        job_name: Optional[str],
        op_exception: Optional[Exception],
        instance: Optional["DagsterInstance"],
    ):
        from dagster._core.execution.build_resources import (
            build_resources,
            wrap_resources_for_execution,
        )
        from dagster._core.execution.context_creation_job import initialize_console_manager

        self._op = None
        if op is not None:

            @graph(name="hook_context_container")
            def temp_graph():
                op()

            self._op = temp_graph.nodes[0]

        # Open resource context manager
        self._resource_defs = wrap_resources_for_execution(resources)
        self._resources_cm = build_resources(self._resource_defs)
        self._resources = self._resources_cm.__enter__()
        self._resources_contain_cm = isinstance(self._resources, IContainsGenerator)

        self._run_id = run_id
        self._job_name = job_name
        self._op_exception = op_exception
        self._instance = instance

        self._log = initialize_console_manager(None)

        self._cm_scope_entered = False

    def __enter__(self):
        self._cm_scope_entered = True
        return self

    def __exit__(self, *exc: Any):
        self._resources_cm.__exit__(*exc)

    def __del__(self):
        if self._resources_contain_cm and not self._cm_scope_entered:
            self._resources_cm.__exit__(None, None, None)

    @property
    def job_name(self) -> str:
        return _check_property_on_test_context(
            self, attr_str="_job_name", user_facing_name="job_name", param_on_builder="job_name"
        )

    @property
    def run_id(self) -> str:
        return _check_property_on_test_context(
            self, attr_str="_run_id", user_facing_name="run_id", param_on_builder="run_id"
        )

    @property
    def hook_def(self) -> HookDefinition:
        raise DagsterInvalidPropertyError(_property_msg("hook_def", "property"))

    @property
    def op(self) -> Node:
        return _check_property_on_test_context(
            self, attr_str="_op", user_facing_name="op", param_on_builder="op"
        )

    @property
    def step(self) -> ExecutionStep:
        raise DagsterInvalidPropertyError(_property_msg("step", "property"))

    @property
    def step_key(self) -> str:
        raise DagsterInvalidPropertyError(_property_msg("step_key", "property"))

    @property
    def required_resource_keys(self) -> set[str]:
        raise DagsterInvalidPropertyError(_property_msg("hook_def", "property"))

    @property
    def resources(self) -> "Resources":
        if self._resources_contain_cm and not self._cm_scope_entered:
            raise DagsterInvariantViolationError(
                "At least one provided resource is a generator, but attempting to access "
                "resources outside of context manager scope. You can use the following syntax to "
                "open a context manager: `with build_hook_context(...) as context:`"
            )
        return self._resources

    @property
    def op_config(self) -> Any:
        raise DagsterInvalidPropertyError(_property_msg("op_config", "property"))

    @property
    def log(self) -> DagsterLogManager:
        return self._log

    @property
    def op_exception(self) -> Optional[BaseException]:
        return self._op_exception

    @property
    def op_output_values(self) -> Mapping[str, Union[Any, Mapping[str, Any]]]:
        """The computed output values.

        Returns a dictionary where keys are output names and the values are:
            * the output values in the normal case
            * a dictionary from mapping key to corresponding value in the mapped case
        """
        raise DagsterInvalidPropertyError(_property_msg("op_output_values", "method"))

    @property
    def op_output_metadata(self) -> Mapping[str, Union[Any, Mapping[str, Any]]]:
        """The applied output metadata.

        Returns a dictionary where keys are output names and the values are:
            * the applied output metadata in the normal case
            * a dictionary from mapping key to corresponding metadata in the mapped case
        """
        raise DagsterInvalidPropertyError(_property_msg("op_output_metadata", "method"))

    @property
    def instance(self) -> "DagsterInstance":
        if not self._instance:
            raise DagsterInvariantViolationError(
                "Tried to access the HookContext instance, but no instance was provided to"
                " `build_hook_context`."
            )

        return self._instance


class BoundHookContext(HookContext):
    def __init__(
        self,
        hook_def: HookDefinition,
        resources: Resources,
        op: Optional[Node],
        log_manager: DagsterLogManager,
        run_id: Optional[str],
        job_name: Optional[str],
        op_exception: Optional[Exception],
        instance: Optional["DagsterInstance"],
    ):
        self._hook_def = hook_def
        self._resources = resources
        self._op = op
        self._log_manager = log_manager
        self._run_id = run_id
        self._job_name = job_name
        self._op_exception = op_exception
        self._instance = instance

    @property
    def job_name(self) -> str:
        return _check_property_on_test_context(
            self, attr_str="_job_name", user_facing_name="job_name", param_on_builder="job_name"
        )

    @property
    def run_id(self) -> str:
        return _check_property_on_test_context(
            self, attr_str="_run_id", user_facing_name="run_id", param_on_builder="run_id"
        )

    @property
    def hook_def(self) -> HookDefinition:
        return self._hook_def

    @property
    def op(self) -> Node:
        return _check_property_on_test_context(
            self, attr_str="_op", user_facing_name="op", param_on_builder="op"
        )

    @property
    def step(self) -> ExecutionStep:
        raise DagsterInvalidPropertyError(_property_msg("step", "property"))

    @property
    def step_key(self) -> str:
        raise DagsterInvalidPropertyError(_property_msg("step_key", "property"))

    @property
    def required_resource_keys(self) -> AbstractSet[str]:
        return self._hook_def.required_resource_keys

    @property
    def resources(self) -> "Resources":
        return self._resources

    @property
    def op_config(self) -> Any:
        raise DagsterInvalidPropertyError(_property_msg("op_config", "property"))

    @property
    def log(self) -> DagsterLogManager:
        return self._log_manager

    @property
    def op_exception(self):
        return self._op_exception

    @property
    def op_output_values(self) -> Mapping[str, Union[Any, Mapping[str, Any]]]:
        """The computed output values.

        Returns a dictionary where keys are output names and the values are:
            * the output values in the normal case
            * a dictionary from mapping key to corresponding value in the mapped case
        """
        raise DagsterInvalidPropertyError(_property_msg("op_output_values", "method"))

    @property
    def op_output_metadata(self) -> Mapping[str, Union[Any, Mapping[str, Any]]]:
        """The applied output metadata.

        Returns a dictionary where keys are output names and the values are:
            * the applied output metadata in the normal case
            * a dictionary from mapping key to corresponding metadata in the mapped case
        """
        raise DagsterInvalidPropertyError(_property_msg("op_output_metadata", "method"))

    @property
    def instance(self) -> "DagsterInstance":
        if not self._instance:
            raise DagsterInvariantViolationError(
                "Tried to access the HookContext instance, but no instance was provided to"
                " `build_hook_context`."
            )

        return self._instance


def build_hook_context(
    resources: Optional[Mapping[str, Any]] = None,
    op: Optional[Union[OpDefinition, PendingNodeInvocation]] = None,
    run_id: Optional[str] = None,
    job_name: Optional[str] = None,
    op_exception: Optional[Exception] = None,
    instance: Optional["DagsterInstance"] = None,
) -> UnboundHookContext:
    """Builds hook context from provided parameters.

    ``build_hook_context`` can be used as either a function or a context manager. If there is a
    provided resource to ``build_hook_context`` that is a context manager, then it must be used as a
    context manager. This function can be used to provide the context argument to the invocation of
    a hook definition.

    Args:
        resources (Optional[Dict[str, Any]]): The resources to provide to the context. These can
            either be values or resource definitions.
        op (Optional[OpDefinition, PendingNodeInvocation]): The op definition which the
            hook may be associated with.
        run_id (Optional[str]): The id of the run in which the hook is invoked (provided for mocking purposes).
        job_name (Optional[str]): The name of the job in which the hook is used (provided for mocking purposes).
        op_exception (Optional[Exception]): The exception that caused the hook to be triggered.
        instance (Optional[DagsterInstance]): The Dagster instance configured to run the hook.

    Examples:
        .. code-block:: python

            context = build_hook_context()
            hook_to_invoke(context)

            with build_hook_context(resources={"foo": context_manager_resource}) as context:
                hook_to_invoke(context)
    """
    op = check.opt_inst_param(op, "op", (OpDefinition, PendingNodeInvocation))

    from dagster._core.instance import DagsterInstance

    return UnboundHookContext(
        resources=check.opt_mapping_param(resources, "resources", key_type=str),
        op=op,
        run_id=check.opt_str_param(run_id, "run_id"),
        job_name=check.opt_str_param(job_name, "job_name"),
        op_exception=check.opt_inst_param(op_exception, "op_exception", Exception),
        instance=check.opt_inst_param(instance, "instance", DagsterInstance),
    )



from collections.abc import Mapping
from typing import Optional

from dagster._core.definitions.asset_check_spec import AssetCheckSpec
from dagster._core.definitions.assets import AssetsDefinition
from dagster._core.definitions.events import AssetKey
from dagster._core.definitions.op_definition import OpDefinition
from dagster._core.definitions.resource_definition import ResourceDefinition


class AssetChecksDefinition(AssetsDefinition):
    """Defines a set of checks that are produced by the same op or op graph.

    AssetChecksDefinition should not be instantiated directly, but rather produced using the `@asset_check` decorator or `AssetChecksDefinition.create` method.
    """

    @staticmethod
    def create(
        *,
        keys_by_input_name: Mapping[str, AssetKey],
        node_def: OpDefinition,
        check_specs_by_output_name: Mapping[str, AssetCheckSpec],
        can_subset: bool,
        resource_defs: Optional[Mapping[str, ResourceDefinition]] = None,
    ):
        """Create an AssetChecksDefinition."""
        return AssetChecksDefinition(
            keys_by_input_name=keys_by_input_name,
            keys_by_output_name={},
            node_def=node_def,
            partitions_def=None,
            partition_mappings=None,
            asset_deps=None,
            selected_asset_keys=None,
            can_subset=can_subset,
            resource_defs=resource_defs,
            group_names_by_key=None,
            metadata_by_key=None,
            tags_by_key=None,
            freshness_policies_by_key=None,
            backfill_policy=None,
            descriptions_by_key=None,
            check_specs_by_output_name=check_specs_by_output_name,
            selected_asset_check_keys=None,
            is_subset=False,
            owners_by_key=None,
        )


# This is still needed in a few places where we need to handle normal AssetsDefinition and
# AssetChecksDefinition differently, but eventually those areas should be refactored and this should
# be removed.
def has_only_asset_checks(assets_def: AssetsDefinition) -> bool:
    return len(assets_def.keys) == 0 and len(assets_def.check_keys) > 0



# flake8: noqa
# type: ignore
# Generated from AssetSelection.g4 by ANTLR 4.13.2
# encoding: utf-8
from antlr4 import *
from io import StringIO
import sys
from typing import TextIO


def serializedATN():
    return [
        4,
        1,
        27,
        117,
        2,
        0,
        7,
        0,
        2,
        1,
        7,
        1,
        2,
        2,
        7,
        2,
        2,
        3,
        7,
        3,
        2,
        4,
        7,
        4,
        2,
        5,
        7,
        5,
        2,
        6,
        7,
        6,
        2,
        7,
        7,
        7,
        2,
        8,
        7,
        8,
        1,
        0,
        1,
        0,
        1,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        3,
        1,
        37,
        8,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        5,
        1,
        45,
        8,
        1,
        10,
        1,
        12,
        1,
        48,
        9,
        1,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        3,
        2,
        60,
        8,
        2,
        1,
        3,
        3,
        3,
        63,
        8,
        3,
        1,
        3,
        1,
        3,
        1,
        4,
        1,
        4,
        3,
        4,
        69,
        8,
        4,
        1,
        5,
        1,
        5,
        1,
        6,
        1,
        6,
        1,
        6,
        1,
        6,
        1,
        6,
        1,
        6,
        1,
        6,
        1,
        6,
        3,
        6,
        81,
        8,
        6,
        1,
        6,
        1,
        6,
        1,
        6,
        1,
        6,
        1,
        6,
        1,
        6,
        1,
        6,
        1,
        6,
        1,
        6,
        1,
        6,
        1,
        6,
        1,
        6,
        1,
        6,
        1,
        6,
        1,
        6,
        1,
        6,
        1,
        6,
        1,
        6,
        1,
        6,
        1,
        6,
        3,
        6,
        103,
        8,
        6,
        1,
        6,
        1,
        6,
        1,
        6,
        1,
        6,
        1,
        6,
        1,
        6,
        3,
        6,
        111,
        8,
        6,
        1,
        7,
        1,
        7,
        1,
        8,
        1,
        8,
        1,
        8,
        0,
        1,
        2,
        9,
        0,
        2,
        4,
        6,
        8,
        10,
        12,
        14,
        16,
        0,
        3,
        1,
        0,
        22,
        23,
        1,
        0,
        24,
        25,
        1,
        0,
        24,
        26,
        129,
        0,
        18,
        1,
        0,
        0,
        0,
        2,
        36,
        1,
        0,
        0,
        0,
        4,
        59,
        1,
        0,
        0,
        0,
        6,
        62,
        1,
        0,
        0,
        0,
        8,
        66,
        1,
        0,
        0,
        0,
        10,
        70,
        1,
        0,
        0,
        0,
        12,
        110,
        1,
        0,
        0,
        0,
        14,
        112,
        1,
        0,
        0,
        0,
        16,
        114,
        1,
        0,
        0,
        0,
        18,
        19,
        3,
        2,
        1,
        0,
        19,
        20,
        5,
        0,
        0,
        1,
        20,
        1,
        1,
        0,
        0,
        0,
        21,
        22,
        6,
        1,
        -1,
        0,
        22,
        37,
        3,
        4,
        2,
        0,
        23,
        24,
        3,
        6,
        3,
        0,
        24,
        25,
        3,
        4,
        2,
        0,
        25,
        26,
        3,
        8,
        4,
        0,
        26,
        37,
        1,
        0,
        0,
        0,
        27,
        28,
        3,
        6,
        3,
        0,
        28,
        29,
        3,
        4,
        2,
        0,
        29,
        37,
        1,
        0,
        0,
        0,
        30,
        31,
        3,
        4,
        2,
        0,
        31,
        32,
        3,
        8,
        4,
        0,
        32,
        37,
        1,
        0,
        0,
        0,
        33,
        34,
        5,
        4,
        0,
        0,
        34,
        37,
        3,
        2,
        1,
        4,
        35,
        37,
        5,
        5,
        0,
        0,
        36,
        21,
        1,
        0,
        0,
        0,
        36,
        23,
        1,
        0,
        0,
        0,
        36,
        27,
        1,
        0,
        0,
        0,
        36,
        30,
        1,
        0,
        0,
        0,
        36,
        33,
        1,
        0,
        0,
        0,
        36,
        35,
        1,
        0,
        0,
        0,
        37,
        46,
        1,
        0,
        0,
        0,
        38,
        39,
        10,
        3,
        0,
        0,
        39,
        40,
        5,
        2,
        0,
        0,
        40,
        45,
        3,
        2,
        1,
        4,
        41,
        42,
        10,
        2,
        0,
        0,
        42,
        43,
        5,
        3,
        0,
        0,
        43,
        45,
        3,
        2,
        1,
        3,
        44,
        38,
        1,
        0,
        0,
        0,
        44,
        41,
        1,
        0,
        0,
        0,
        45,
        48,
        1,
        0,
        0,
        0,
        46,
        44,
        1,
        0,
        0,
        0,
        46,
        47,
        1,
        0,
        0,
        0,
        47,
        3,
        1,
        0,
        0,
        0,
        48,
        46,
        1,
        0,
        0,
        0,
        49,
        60,
        3,
        12,
        6,
        0,
        50,
        51,
        3,
        10,
        5,
        0,
        51,
        52,
        5,
        9,
        0,
        0,
        52,
        53,
        3,
        2,
        1,
        0,
        53,
        54,
        5,
        10,
        0,
        0,
        54,
        60,
        1,
        0,
        0,
        0,
        55,
        56,
        5,
        9,
        0,
        0,
        56,
        57,
        3,
        2,
        1,
        0,
        57,
        58,
        5,
        10,
        0,
        0,
        58,
        60,
        1,
        0,
        0,
        0,
        59,
        49,
        1,
        0,
        0,
        0,
        59,
        50,
        1,
        0,
        0,
        0,
        59,
        55,
        1,
        0,
        0,
        0,
        60,
        5,
        1,
        0,
        0,
        0,
        61,
        63,
        5,
        7,
        0,
        0,
        62,
        61,
        1,
        0,
        0,
        0,
        62,
        63,
        1,
        0,
        0,
        0,
        63,
        64,
        1,
        0,
        0,
        0,
        64,
        65,
        5,
        6,
        0,
        0,
        65,
        7,
        1,
        0,
        0,
        0,
        66,
        68,
        5,
        6,
        0,
        0,
        67,
        69,
        5,
        7,
        0,
        0,
        68,
        67,
        1,
        0,
        0,
        0,
        68,
        69,
        1,
        0,
        0,
        0,
        69,
        9,
        1,
        0,
        0,
        0,
        70,
        71,
        7,
        0,
        0,
        0,
        71,
        11,
        1,
        0,
        0,
        0,
        72,
        73,
        5,
        12,
        0,
        0,
        73,
        74,
        5,
        8,
        0,
        0,
        74,
        111,
        3,
        16,
        8,
        0,
        75,
        76,
        5,
        15,
        0,
        0,
        76,
        77,
        5,
        8,
        0,
        0,
        77,
        80,
        3,
        14,
        7,
        0,
        78,
        79,
        5,
        1,
        0,
        0,
        79,
        81,
        3,
        14,
        7,
        0,
        80,
        78,
        1,
        0,
        0,
        0,
        80,
        81,
        1,
        0,
        0,
        0,
        81,
        111,
        1,
        0,
        0,
        0,
        82,
        83,
        5,
        13,
        0,
        0,
        83,
        84,
        5,
        8,
        0,
        0,
        84,
        111,
        3,
        14,
        7,
        0,
        85,
        86,
        5,
        14,
        0,
        0,
        86,
        87,
        5,
        8,
        0,
        0,
        87,
        111,
        3,
        14,
        7,
        0,
        88,
        89,
        5,
        16,
        0,
        0,
        89,
        90,
        5,
        8,
        0,
        0,
        90,
        111,
        3,
        14,
        7,
        0,
        91,
        92,
        5,
        18,
        0,
        0,
        92,
        93,
        5,
        8,
        0,
        0,
        93,
        111,
        3,
        14,
        7,
        0,
        94,
        95,
        5,
        19,
        0,
        0,
        95,
        96,
        5,
        8,
        0,
        0,
        96,
        111,
        3,
        14,
        7,
        0,
        97,
        98,
        5,
        20,
        0,
        0,
        98,
        99,
        5,
        8,
        0,
        0,
        99,
        102,
        3,
        14,
        7,
        0,
        100,
        101,
        5,
        1,
        0,
        0,
        101,
        103,
        3,
        14,
        7,
        0,
        102,
        100,
        1,
        0,
        0,
        0,
        102,
        103,
        1,
        0,
        0,
        0,
        103,
        111,
        1,
        0,
        0,
        0,
        104,
        105,
        5,
        17,
        0,
        0,
        105,
        106,
        5,
        8,
        0,
        0,
        106,
        111,
        3,
        14,
        7,
        0,
        107,
        108,
        5,
        21,
        0,
        0,
        108,
        109,
        5,
        8,
        0,
        0,
        109,
        111,
        3,
        14,
        7,
        0,
        110,
        72,
        1,
        0,
        0,
        0,
        110,
        75,
        1,
        0,
        0,
        0,
        110,
        82,
        1,
        0,
        0,
        0,
        110,
        85,
        1,
        0,
        0,
        0,
        110,
        88,
        1,
        0,
        0,
        0,
        110,
        91,
        1,
        0,
        0,
        0,
        110,
        94,
        1,
        0,
        0,
        0,
        110,
        97,
        1,
        0,
        0,
        0,
        110,
        104,
        1,
        0,
        0,
        0,
        110,
        107,
        1,
        0,
        0,
        0,
        111,
        13,
        1,
        0,
        0,
        0,
        112,
        113,
        7,
        1,
        0,
        0,
        113,
        15,
        1,
        0,
        0,
        0,
        114,
        115,
        7,
        2,
        0,
        0,
        115,
        17,
        1,
        0,
        0,
        0,
        9,
        36,
        44,
        46,
        59,
        62,
        68,
        80,
        102,
        110,
    ]


class AssetSelectionParser(Parser):
    grammarFileName = "AssetSelection.g4"

    atn = ATNDeserializer().deserialize(serializedATN())

    decisionsToDFA = [DFA(ds, i) for i, ds in enumerate(atn.decisionToState)]

    sharedContextCache = PredictionContextCache()

    literalNames = [
        "<INVALID>",
        "'='",
        "'and'",
        "'or'",
        "'not'",
        "'*'",
        "'+'",
        "<INVALID>",
        "':'",
        "'('",
        "')'",
        "','",
        "'key'",
        "'owner'",
        "'group'",
        "'tag'",
        "'kind'",
        "'code_location'",
        "'column'",
        "'table_name'",
        "'column_tag'",
        "'changed_in_branch'",
        "'sinks'",
        "'roots'",
    ]

    symbolicNames = [
        "<INVALID>",
        "EQUAL",
        "AND",
        "OR",
        "NOT",
        "STAR",
        "PLUS",
        "DIGITS",
        "COLON",
        "LPAREN",
        "RPAREN",
        "COMMA",
        "KEY",
        "OWNER",
        "GROUP",
        "TAG",
        "KIND",
        "CODE_LOCATION",
        "COLUMN",
        "TABLE_NAME",
        "COLUMN_TAG",
        "CHANGED_IN_BRANCH",
        "SINKS",
        "ROOTS",
        "QUOTED_STRING",
        "UNQUOTED_STRING",
        "UNQUOTED_WILDCARD_STRING",
        "WS",
    ]

    RULE_start = 0
    RULE_expr = 1
    RULE_traversalAllowedExpr = 2
    RULE_upTraversal = 3
    RULE_downTraversal = 4
    RULE_functionName = 5
    RULE_attributeExpr = 6
    RULE_value = 7
    RULE_keyValue = 8

    ruleNames = [
        "start",
        "expr",
        "traversalAllowedExpr",
        "upTraversal",
        "downTraversal",
        "functionName",
        "attributeExpr",
        "value",
        "keyValue",
    ]

    EOF = Token.EOF
    EQUAL = 1
    AND = 2
    OR = 3
    NOT = 4
    STAR = 5
    PLUS = 6
    DIGITS = 7
    COLON = 8
    LPAREN = 9
    RPAREN = 10
    COMMA = 11
    KEY = 12
    OWNER = 13
    GROUP = 14
    TAG = 15
    KIND = 16
    CODE_LOCATION = 17
    COLUMN = 18
    TABLE_NAME = 19
    COLUMN_TAG = 20
    CHANGED_IN_BRANCH = 21
    SINKS = 22
    ROOTS = 23
    QUOTED_STRING = 24
    UNQUOTED_STRING = 25
    UNQUOTED_WILDCARD_STRING = 26
    WS = 27

    def __init__(self, input: TokenStream, output: TextIO = sys.stdout):
        super().__init__(input, output)
        self.checkVersion("4.13.2")
        self._interp = ParserATNSimulator(
            self, self.atn, self.decisionsToDFA, self.sharedContextCache
        )
        self._predicates = None

    class StartContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def expr(self):
            return self.getTypedRuleContext(AssetSelectionParser.ExprContext, 0)

        def EOF(self):
            return self.getToken(AssetSelectionParser.EOF, 0)

        def getRuleIndex(self):
            return AssetSelectionParser.RULE_start

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterStart"):
                listener.enterStart(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitStart"):
                listener.exitStart(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitStart"):
                return visitor.visitStart(self)
            else:
                return visitor.visitChildren(self)

    def start(self):
        localctx = AssetSelectionParser.StartContext(self, self._ctx, self.state)
        self.enterRule(localctx, 0, self.RULE_start)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 18
            self.expr(0)
            self.state = 19
            self.match(AssetSelectionParser.EOF)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class ExprContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def getRuleIndex(self):
            return AssetSelectionParser.RULE_expr

        def copyFrom(self, ctx: ParserRuleContext):
            super().copyFrom(ctx)

    class UpTraversalExpressionContext(ExprContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a AssetSelectionParser.ExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def upTraversal(self):
            return self.getTypedRuleContext(AssetSelectionParser.UpTraversalContext, 0)

        def traversalAllowedExpr(self):
            return self.getTypedRuleContext(AssetSelectionParser.TraversalAllowedExprContext, 0)

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterUpTraversalExpression"):
                listener.enterUpTraversalExpression(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitUpTraversalExpression"):
                listener.exitUpTraversalExpression(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitUpTraversalExpression"):
                return visitor.visitUpTraversalExpression(self)
            else:
                return visitor.visitChildren(self)

    class AndExpressionContext(ExprContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a AssetSelectionParser.ExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def expr(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(AssetSelectionParser.ExprContext)
            else:
                return self.getTypedRuleContext(AssetSelectionParser.ExprContext, i)

        def AND(self):
            return self.getToken(AssetSelectionParser.AND, 0)

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterAndExpression"):
                listener.enterAndExpression(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitAndExpression"):
                listener.exitAndExpression(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitAndExpression"):
                return visitor.visitAndExpression(self)
            else:
                return visitor.visitChildren(self)

    class AllExpressionContext(ExprContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a AssetSelectionParser.ExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def STAR(self):
            return self.getToken(AssetSelectionParser.STAR, 0)

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterAllExpression"):
                listener.enterAllExpression(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitAllExpression"):
                listener.exitAllExpression(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitAllExpression"):
                return visitor.visitAllExpression(self)
            else:
                return visitor.visitChildren(self)

    class TraversalAllowedExpressionContext(ExprContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a AssetSelectionParser.ExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def traversalAllowedExpr(self):
            return self.getTypedRuleContext(AssetSelectionParser.TraversalAllowedExprContext, 0)

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterTraversalAllowedExpression"):
                listener.enterTraversalAllowedExpression(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitTraversalAllowedExpression"):
                listener.exitTraversalAllowedExpression(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitTraversalAllowedExpression"):
                return visitor.visitTraversalAllowedExpression(self)
            else:
                return visitor.visitChildren(self)

    class DownTraversalExpressionContext(ExprContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a AssetSelectionParser.ExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def traversalAllowedExpr(self):
            return self.getTypedRuleContext(AssetSelectionParser.TraversalAllowedExprContext, 0)

        def downTraversal(self):
            return self.getTypedRuleContext(AssetSelectionParser.DownTraversalContext, 0)

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterDownTraversalExpression"):
                listener.enterDownTraversalExpression(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitDownTraversalExpression"):
                listener.exitDownTraversalExpression(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitDownTraversalExpression"):
                return visitor.visitDownTraversalExpression(self)
            else:
                return visitor.visitChildren(self)

    class NotExpressionContext(ExprContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a AssetSelectionParser.ExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def NOT(self):
            return self.getToken(AssetSelectionParser.NOT, 0)

        def expr(self):
            return self.getTypedRuleContext(AssetSelectionParser.ExprContext, 0)

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterNotExpression"):
                listener.enterNotExpression(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitNotExpression"):
                listener.exitNotExpression(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitNotExpression"):
                return visitor.visitNotExpression(self)
            else:
                return visitor.visitChildren(self)

    class OrExpressionContext(ExprContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a AssetSelectionParser.ExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def expr(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(AssetSelectionParser.ExprContext)
            else:
                return self.getTypedRuleContext(AssetSelectionParser.ExprContext, i)

        def OR(self):
            return self.getToken(AssetSelectionParser.OR, 0)

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterOrExpression"):
                listener.enterOrExpression(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitOrExpression"):
                listener.exitOrExpression(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitOrExpression"):
                return visitor.visitOrExpression(self)
            else:
                return visitor.visitChildren(self)

    class UpAndDownTraversalExpressionContext(ExprContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a AssetSelectionParser.ExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def upTraversal(self):
            return self.getTypedRuleContext(AssetSelectionParser.UpTraversalContext, 0)

        def traversalAllowedExpr(self):
            return self.getTypedRuleContext(AssetSelectionParser.TraversalAllowedExprContext, 0)

        def downTraversal(self):
            return self.getTypedRuleContext(AssetSelectionParser.DownTraversalContext, 0)

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterUpAndDownTraversalExpression"):
                listener.enterUpAndDownTraversalExpression(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitUpAndDownTraversalExpression"):
                listener.exitUpAndDownTraversalExpression(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitUpAndDownTraversalExpression"):
                return visitor.visitUpAndDownTraversalExpression(self)
            else:
                return visitor.visitChildren(self)

    def expr(self, _p: int = 0):
        _parentctx = self._ctx
        _parentState = self.state
        localctx = AssetSelectionParser.ExprContext(self, self._ctx, _parentState)
        _prevctx = localctx
        _startState = 2
        self.enterRecursionRule(localctx, 2, self.RULE_expr, _p)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 36
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 0, self._ctx)
            if la_ == 1:
                localctx = AssetSelectionParser.TraversalAllowedExpressionContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx

                self.state = 22
                self.traversalAllowedExpr()
                pass

            elif la_ == 2:
                localctx = AssetSelectionParser.UpAndDownTraversalExpressionContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 23
                self.upTraversal()
                self.state = 24
                self.traversalAllowedExpr()
                self.state = 25
                self.downTraversal()
                pass

            elif la_ == 3:
                localctx = AssetSelectionParser.UpTraversalExpressionContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 27
                self.upTraversal()
                self.state = 28
                self.traversalAllowedExpr()
                pass

            elif la_ == 4:
                localctx = AssetSelectionParser.DownTraversalExpressionContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 30
                self.traversalAllowedExpr()
                self.state = 31
                self.downTraversal()
                pass

            elif la_ == 5:
                localctx = AssetSelectionParser.NotExpressionContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 33
                self.match(AssetSelectionParser.NOT)
                self.state = 34
                self.expr(4)
                pass

            elif la_ == 6:
                localctx = AssetSelectionParser.AllExpressionContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 35
                self.match(AssetSelectionParser.STAR)
                pass

            self._ctx.stop = self._input.LT(-1)
            self.state = 46
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input, 2, self._ctx)
            while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    if self._parseListeners is not None:
                        self.triggerExitRuleEvent()
                    _prevctx = localctx
                    self.state = 44
                    self._errHandler.sync(self)
                    la_ = self._interp.adaptivePredict(self._input, 1, self._ctx)
                    if la_ == 1:
                        localctx = AssetSelectionParser.AndExpressionContext(
                            self, AssetSelectionParser.ExprContext(self, _parentctx, _parentState)
                        )
                        self.pushNewRecursionContext(localctx, _startState, self.RULE_expr)
                        self.state = 38
                        if not self.precpred(self._ctx, 3):
                            from antlr4.error.Errors import FailedPredicateException

                            raise FailedPredicateException(self, "self.precpred(self._ctx, 3)")
                        self.state = 39
                        self.match(AssetSelectionParser.AND)
                        self.state = 40
                        self.expr(4)
                        pass

                    elif la_ == 2:
                        localctx = AssetSelectionParser.OrExpressionContext(
                            self, AssetSelectionParser.ExprContext(self, _parentctx, _parentState)
                        )
                        self.pushNewRecursionContext(localctx, _startState, self.RULE_expr)
                        self.state = 41
                        if not self.precpred(self._ctx, 2):
                            from antlr4.error.Errors import FailedPredicateException

                            raise FailedPredicateException(self, "self.precpred(self._ctx, 2)")
                        self.state = 42
                        self.match(AssetSelectionParser.OR)
                        self.state = 43
                        self.expr(3)
                        pass

                self.state = 48
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 2, self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.unrollRecursionContexts(_parentctx)
        return localctx

    class TraversalAllowedExprContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def getRuleIndex(self):
            return AssetSelectionParser.RULE_traversalAllowedExpr

        def copyFrom(self, ctx: ParserRuleContext):
            super().copyFrom(ctx)

    class ParenthesizedExpressionContext(TraversalAllowedExprContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a AssetSelectionParser.TraversalAllowedExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def LPAREN(self):
            return self.getToken(AssetSelectionParser.LPAREN, 0)

        def expr(self):
            return self.getTypedRuleContext(AssetSelectionParser.ExprContext, 0)

        def RPAREN(self):
            return self.getToken(AssetSelectionParser.RPAREN, 0)

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterParenthesizedExpression"):
                listener.enterParenthesizedExpression(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitParenthesizedExpression"):
                listener.exitParenthesizedExpression(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitParenthesizedExpression"):
                return visitor.visitParenthesizedExpression(self)
            else:
                return visitor.visitChildren(self)

    class AttributeExpressionContext(TraversalAllowedExprContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a AssetSelectionParser.TraversalAllowedExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def attributeExpr(self):
            return self.getTypedRuleContext(AssetSelectionParser.AttributeExprContext, 0)

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterAttributeExpression"):
                listener.enterAttributeExpression(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitAttributeExpression"):
                listener.exitAttributeExpression(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitAttributeExpression"):
                return visitor.visitAttributeExpression(self)
            else:
                return visitor.visitChildren(self)

    class FunctionCallExpressionContext(TraversalAllowedExprContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a AssetSelectionParser.TraversalAllowedExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def functionName(self):
            return self.getTypedRuleContext(AssetSelectionParser.FunctionNameContext, 0)

        def LPAREN(self):
            return self.getToken(AssetSelectionParser.LPAREN, 0)

        def expr(self):
            return self.getTypedRuleContext(AssetSelectionParser.ExprContext, 0)

        def RPAREN(self):
            return self.getToken(AssetSelectionParser.RPAREN, 0)

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterFunctionCallExpression"):
                listener.enterFunctionCallExpression(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitFunctionCallExpression"):
                listener.exitFunctionCallExpression(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitFunctionCallExpression"):
                return visitor.visitFunctionCallExpression(self)
            else:
                return visitor.visitChildren(self)

    def traversalAllowedExpr(self):
        localctx = AssetSelectionParser.TraversalAllowedExprContext(self, self._ctx, self.state)
        self.enterRule(localctx, 4, self.RULE_traversalAllowedExpr)
        try:
            self.state = 59
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [12, 13, 14, 15, 16, 17, 18, 19, 20, 21]:
                localctx = AssetSelectionParser.AttributeExpressionContext(self, localctx)
                self.enterOuterAlt(localctx, 1)
                self.state = 49
                self.attributeExpr()
                pass
            elif token in [22, 23]:
                localctx = AssetSelectionParser.FunctionCallExpressionContext(self, localctx)
                self.enterOuterAlt(localctx, 2)
                self.state = 50
                self.functionName()
                self.state = 51
                self.match(AssetSelectionParser.LPAREN)
                self.state = 52
                self.expr(0)
                self.state = 53
                self.match(AssetSelectionParser.RPAREN)
                pass
            elif token in [9]:
                localctx = AssetSelectionParser.ParenthesizedExpressionContext(self, localctx)
                self.enterOuterAlt(localctx, 3)
                self.state = 55
                self.match(AssetSelectionParser.LPAREN)
                self.state = 56
                self.expr(0)
                self.state = 57
                self.match(AssetSelectionParser.RPAREN)
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class UpTraversalContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def PLUS(self):
            return self.getToken(AssetSelectionParser.PLUS, 0)

        def DIGITS(self):
            return self.getToken(AssetSelectionParser.DIGITS, 0)

        def getRuleIndex(self):
            return AssetSelectionParser.RULE_upTraversal

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterUpTraversal"):
                listener.enterUpTraversal(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitUpTraversal"):
                listener.exitUpTraversal(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitUpTraversal"):
                return visitor.visitUpTraversal(self)
            else:
                return visitor.visitChildren(self)

    def upTraversal(self):
        localctx = AssetSelectionParser.UpTraversalContext(self, self._ctx, self.state)
        self.enterRule(localctx, 6, self.RULE_upTraversal)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 62
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la == 7:
                self.state = 61
                self.match(AssetSelectionParser.DIGITS)

            self.state = 64
            self.match(AssetSelectionParser.PLUS)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class DownTraversalContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def PLUS(self):
            return self.getToken(AssetSelectionParser.PLUS, 0)

        def DIGITS(self):
            return self.getToken(AssetSelectionParser.DIGITS, 0)

        def getRuleIndex(self):
            return AssetSelectionParser.RULE_downTraversal

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterDownTraversal"):
                listener.enterDownTraversal(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitDownTraversal"):
                listener.exitDownTraversal(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitDownTraversal"):
                return visitor.visitDownTraversal(self)
            else:
                return visitor.visitChildren(self)

    def downTraversal(self):
        localctx = AssetSelectionParser.DownTraversalContext(self, self._ctx, self.state)
        self.enterRule(localctx, 8, self.RULE_downTraversal)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 66
            self.match(AssetSelectionParser.PLUS)
            self.state = 68
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 5, self._ctx)
            if la_ == 1:
                self.state = 67
                self.match(AssetSelectionParser.DIGITS)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class FunctionNameContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def SINKS(self):
            return self.getToken(AssetSelectionParser.SINKS, 0)

        def ROOTS(self):
            return self.getToken(AssetSelectionParser.ROOTS, 0)

        def getRuleIndex(self):
            return AssetSelectionParser.RULE_functionName

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterFunctionName"):
                listener.enterFunctionName(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitFunctionName"):
                listener.exitFunctionName(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitFunctionName"):
                return visitor.visitFunctionName(self)
            else:
                return visitor.visitChildren(self)

    def functionName(self):
        localctx = AssetSelectionParser.FunctionNameContext(self, self._ctx, self.state)
        self.enterRule(localctx, 10, self.RULE_functionName)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 70
            _la = self._input.LA(1)
            if not (_la == 22 or _la == 23):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class AttributeExprContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def getRuleIndex(self):
            return AssetSelectionParser.RULE_attributeExpr

        def copyFrom(self, ctx: ParserRuleContext):
            super().copyFrom(ctx)

    class GroupAttributeExprContext(AttributeExprContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a AssetSelectionParser.AttributeExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def GROUP(self):
            return self.getToken(AssetSelectionParser.GROUP, 0)

        def COLON(self):
            return self.getToken(AssetSelectionParser.COLON, 0)

        def value(self):
            return self.getTypedRuleContext(AssetSelectionParser.ValueContext, 0)

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterGroupAttributeExpr"):
                listener.enterGroupAttributeExpr(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitGroupAttributeExpr"):
                listener.exitGroupAttributeExpr(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitGroupAttributeExpr"):
                return visitor.visitGroupAttributeExpr(self)
            else:
                return visitor.visitChildren(self)

    class TagAttributeExprContext(AttributeExprContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a AssetSelectionParser.AttributeExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def TAG(self):
            return self.getToken(AssetSelectionParser.TAG, 0)

        def COLON(self):
            return self.getToken(AssetSelectionParser.COLON, 0)

        def value(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(AssetSelectionParser.ValueContext)
            else:
                return self.getTypedRuleContext(AssetSelectionParser.ValueContext, i)

        def EQUAL(self):
            return self.getToken(AssetSelectionParser.EQUAL, 0)

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterTagAttributeExpr"):
                listener.enterTagAttributeExpr(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitTagAttributeExpr"):
                listener.exitTagAttributeExpr(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitTagAttributeExpr"):
                return visitor.visitTagAttributeExpr(self)
            else:
                return visitor.visitChildren(self)

    class KeyExprContext(AttributeExprContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a AssetSelectionParser.AttributeExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def KEY(self):
            return self.getToken(AssetSelectionParser.KEY, 0)

        def COLON(self):
            return self.getToken(AssetSelectionParser.COLON, 0)

        def keyValue(self):
            return self.getTypedRuleContext(AssetSelectionParser.KeyValueContext, 0)

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterKeyExpr"):
                listener.enterKeyExpr(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitKeyExpr"):
                listener.exitKeyExpr(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitKeyExpr"):
                return visitor.visitKeyExpr(self)
            else:
                return visitor.visitChildren(self)

    class ColumnAttributeExprContext(AttributeExprContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a AssetSelectionParser.AttributeExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def COLUMN(self):
            return self.getToken(AssetSelectionParser.COLUMN, 0)

        def COLON(self):
            return self.getToken(AssetSelectionParser.COLON, 0)

        def value(self):
            return self.getTypedRuleContext(AssetSelectionParser.ValueContext, 0)

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterColumnAttributeExpr"):
                listener.enterColumnAttributeExpr(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitColumnAttributeExpr"):
                listener.exitColumnAttributeExpr(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitColumnAttributeExpr"):
                return visitor.visitColumnAttributeExpr(self)
            else:
                return visitor.visitChildren(self)

    class KindAttributeExprContext(AttributeExprContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a AssetSelectionParser.AttributeExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def KIND(self):
            return self.getToken(AssetSelectionParser.KIND, 0)

        def COLON(self):
            return self.getToken(AssetSelectionParser.COLON, 0)

        def value(self):
            return self.getTypedRuleContext(AssetSelectionParser.ValueContext, 0)

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterKindAttributeExpr"):
                listener.enterKindAttributeExpr(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitKindAttributeExpr"):
                listener.exitKindAttributeExpr(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitKindAttributeExpr"):
                return visitor.visitKindAttributeExpr(self)
            else:
                return visitor.visitChildren(self)

    class CodeLocationAttributeExprContext(AttributeExprContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a AssetSelectionParser.AttributeExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def CODE_LOCATION(self):
            return self.getToken(AssetSelectionParser.CODE_LOCATION, 0)

        def COLON(self):
            return self.getToken(AssetSelectionParser.COLON, 0)

        def value(self):
            return self.getTypedRuleContext(AssetSelectionParser.ValueContext, 0)

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterCodeLocationAttributeExpr"):
                listener.enterCodeLocationAttributeExpr(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitCodeLocationAttributeExpr"):
                listener.exitCodeLocationAttributeExpr(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitCodeLocationAttributeExpr"):
                return visitor.visitCodeLocationAttributeExpr(self)
            else:
                return visitor.visitChildren(self)

    class OwnerAttributeExprContext(AttributeExprContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a AssetSelectionParser.AttributeExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def OWNER(self):
            return self.getToken(AssetSelectionParser.OWNER, 0)

        def COLON(self):
            return self.getToken(AssetSelectionParser.COLON, 0)

        def value(self):
            return self.getTypedRuleContext(AssetSelectionParser.ValueContext, 0)

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterOwnerAttributeExpr"):
                listener.enterOwnerAttributeExpr(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitOwnerAttributeExpr"):
                listener.exitOwnerAttributeExpr(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitOwnerAttributeExpr"):
                return visitor.visitOwnerAttributeExpr(self)
            else:
                return visitor.visitChildren(self)

    class ChangedInBranchAttributeExprContext(AttributeExprContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a AssetSelectionParser.AttributeExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def CHANGED_IN_BRANCH(self):
            return self.getToken(AssetSelectionParser.CHANGED_IN_BRANCH, 0)

        def COLON(self):
            return self.getToken(AssetSelectionParser.COLON, 0)

        def value(self):
            return self.getTypedRuleContext(AssetSelectionParser.ValueContext, 0)

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterChangedInBranchAttributeExpr"):
                listener.enterChangedInBranchAttributeExpr(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitChangedInBranchAttributeExpr"):
                listener.exitChangedInBranchAttributeExpr(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitChangedInBranchAttributeExpr"):
                return visitor.visitChangedInBranchAttributeExpr(self)
            else:
                return visitor.visitChildren(self)

    class ColumnTagAttributeExprContext(AttributeExprContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a AssetSelectionParser.AttributeExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def COLUMN_TAG(self):
            return self.getToken(AssetSelectionParser.COLUMN_TAG, 0)

        def COLON(self):
            return self.getToken(AssetSelectionParser.COLON, 0)

        def value(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(AssetSelectionParser.ValueContext)
            else:
                return self.getTypedRuleContext(AssetSelectionParser.ValueContext, i)

        def EQUAL(self):
            return self.getToken(AssetSelectionParser.EQUAL, 0)

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterColumnTagAttributeExpr"):
                listener.enterColumnTagAttributeExpr(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitColumnTagAttributeExpr"):
                listener.exitColumnTagAttributeExpr(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitColumnTagAttributeExpr"):
                return visitor.visitColumnTagAttributeExpr(self)
            else:
                return visitor.visitChildren(self)

    class TableNameAttributeExprContext(AttributeExprContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a AssetSelectionParser.AttributeExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def TABLE_NAME(self):
            return self.getToken(AssetSelectionParser.TABLE_NAME, 0)

        def COLON(self):
            return self.getToken(AssetSelectionParser.COLON, 0)

        def value(self):
            return self.getTypedRuleContext(AssetSelectionParser.ValueContext, 0)

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterTableNameAttributeExpr"):
                listener.enterTableNameAttributeExpr(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitTableNameAttributeExpr"):
                listener.exitTableNameAttributeExpr(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitTableNameAttributeExpr"):
                return visitor.visitTableNameAttributeExpr(self)
            else:
                return visitor.visitChildren(self)

    def attributeExpr(self):
        localctx = AssetSelectionParser.AttributeExprContext(self, self._ctx, self.state)
        self.enterRule(localctx, 12, self.RULE_attributeExpr)
        try:
            self.state = 110
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [12]:
                localctx = AssetSelectionParser.KeyExprContext(self, localctx)
                self.enterOuterAlt(localctx, 1)
                self.state = 72
                self.match(AssetSelectionParser.KEY)
                self.state = 73
                self.match(AssetSelectionParser.COLON)
                self.state = 74
                self.keyValue()
                pass
            elif token in [15]:
                localctx = AssetSelectionParser.TagAttributeExprContext(self, localctx)
                self.enterOuterAlt(localctx, 2)
                self.state = 75
                self.match(AssetSelectionParser.TAG)
                self.state = 76
                self.match(AssetSelectionParser.COLON)
                self.state = 77
                self.value()
                self.state = 80
                self._errHandler.sync(self)
                la_ = self._interp.adaptivePredict(self._input, 6, self._ctx)
                if la_ == 1:
                    self.state = 78
                    self.match(AssetSelectionParser.EQUAL)
                    self.state = 79
                    self.value()

                pass
            elif token in [13]:
                localctx = AssetSelectionParser.OwnerAttributeExprContext(self, localctx)
                self.enterOuterAlt(localctx, 3)
                self.state = 82
                self.match(AssetSelectionParser.OWNER)
                self.state = 83
                self.match(AssetSelectionParser.COLON)
                self.state = 84
                self.value()
                pass
            elif token in [14]:
                localctx = AssetSelectionParser.GroupAttributeExprContext(self, localctx)
                self.enterOuterAlt(localctx, 4)
                self.state = 85
                self.match(AssetSelectionParser.GROUP)
                self.state = 86
                self.match(AssetSelectionParser.COLON)
                self.state = 87
                self.value()
                pass
            elif token in [16]:
                localctx = AssetSelectionParser.KindAttributeExprContext(self, localctx)
                self.enterOuterAlt(localctx, 5)
                self.state = 88
                self.match(AssetSelectionParser.KIND)
                self.state = 89
                self.match(AssetSelectionParser.COLON)
                self.state = 90
                self.value()
                pass
            elif token in [18]:
                localctx = AssetSelectionParser.ColumnAttributeExprContext(self, localctx)
                self.enterOuterAlt(localctx, 6)
                self.state = 91
                self.match(AssetSelectionParser.COLUMN)
                self.state = 92
                self.match(AssetSelectionParser.COLON)
                self.state = 93
                self.value()
                pass
            elif token in [19]:
                localctx = AssetSelectionParser.TableNameAttributeExprContext(self, localctx)
                self.enterOuterAlt(localctx, 7)
                self.state = 94
                self.match(AssetSelectionParser.TABLE_NAME)
                self.state = 95
                self.match(AssetSelectionParser.COLON)
                self.state = 96
                self.value()
                pass
            elif token in [20]:
                localctx = AssetSelectionParser.ColumnTagAttributeExprContext(self, localctx)
                self.enterOuterAlt(localctx, 8)
                self.state = 97
                self.match(AssetSelectionParser.COLUMN_TAG)
                self.state = 98
                self.match(AssetSelectionParser.COLON)
                self.state = 99
                self.value()
                self.state = 102
                self._errHandler.sync(self)
                la_ = self._interp.adaptivePredict(self._input, 7, self._ctx)
                if la_ == 1:
                    self.state = 100
                    self.match(AssetSelectionParser.EQUAL)
                    self.state = 101
                    self.value()

                pass
            elif token in [17]:
                localctx = AssetSelectionParser.CodeLocationAttributeExprContext(self, localctx)
                self.enterOuterAlt(localctx, 9)
                self.state = 104
                self.match(AssetSelectionParser.CODE_LOCATION)
                self.state = 105
                self.match(AssetSelectionParser.COLON)
                self.state = 106
                self.value()
                pass
            elif token in [21]:
                localctx = AssetSelectionParser.ChangedInBranchAttributeExprContext(self, localctx)
                self.enterOuterAlt(localctx, 10)
                self.state = 107
                self.match(AssetSelectionParser.CHANGED_IN_BRANCH)
                self.state = 108
                self.match(AssetSelectionParser.COLON)
                self.state = 109
                self.value()
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class ValueContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def QUOTED_STRING(self):
            return self.getToken(AssetSelectionParser.QUOTED_STRING, 0)

        def UNQUOTED_STRING(self):
            return self.getToken(AssetSelectionParser.UNQUOTED_STRING, 0)

        def getRuleIndex(self):
            return AssetSelectionParser.RULE_value

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterValue"):
                listener.enterValue(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitValue"):
                listener.exitValue(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitValue"):
                return visitor.visitValue(self)
            else:
                return visitor.visitChildren(self)

    def value(self):
        localctx = AssetSelectionParser.ValueContext(self, self._ctx, self.state)
        self.enterRule(localctx, 14, self.RULE_value)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 112
            _la = self._input.LA(1)
            if not (_la == 24 or _la == 25):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class KeyValueContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def QUOTED_STRING(self):
            return self.getToken(AssetSelectionParser.QUOTED_STRING, 0)

        def UNQUOTED_STRING(self):
            return self.getToken(AssetSelectionParser.UNQUOTED_STRING, 0)

        def UNQUOTED_WILDCARD_STRING(self):
            return self.getToken(AssetSelectionParser.UNQUOTED_WILDCARD_STRING, 0)

        def getRuleIndex(self):
            return AssetSelectionParser.RULE_keyValue

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterKeyValue"):
                listener.enterKeyValue(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitKeyValue"):
                listener.exitKeyValue(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitKeyValue"):
                return visitor.visitKeyValue(self)
            else:
                return visitor.visitChildren(self)

    def keyValue(self):
        localctx = AssetSelectionParser.KeyValueContext(self, self._ctx, self.state)
        self.enterRule(localctx, 16, self.RULE_keyValue)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 114
            _la = self._input.LA(1)
            if not (((_la) & ~0x3F) == 0 and ((1 << _la) & 117440512) != 0):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    def sempred(self, localctx: RuleContext, ruleIndex: int, predIndex: int):
        if self._predicates == None:
            self._predicates = dict()
        self._predicates[1] = self.expr_sempred
        pred = self._predicates.get(ruleIndex, None)
        if pred is None:
            raise Exception("No predicate with index:" + str(ruleIndex))
        else:
            return pred(localctx, predIndex)

    def expr_sempred(self, localctx: ExprContext, predIndex: int):
        if predIndex == 0:
            return self.precpred(self._ctx, 3)

        if predIndex == 1:
            return self.precpred(self._ctx, 2)



import dagster._check as check
from dagster._config.config_type import ConfigType
from dagster._config.field import Field
from dagster._config.snap import ConfigFieldSnap, ConfigSchemaSnapshot, ConfigTypeSnap
from dagster._config.stack import EvaluationStackEntry
from dagster._record import record


@record
class ContextData:
    config_schema_snapshot: ConfigSchemaSnapshot
    config_type_snap: ConfigTypeSnap
    stack: EvaluationStackEntry

    @property
    def config_type_key(self) -> str:
        return self.config_type_snap.key


@record
class ValidationContext(ContextData):
    def for_field_snap(self, field_snap: ConfigFieldSnap) -> "ValidationContext":
        check.inst_param(field_snap, "field_snap", ConfigFieldSnap)
        field_snap_name = check.not_none(field_snap.name)
        return ValidationContext(
            config_schema_snapshot=self.config_schema_snapshot,
            config_type_snap=self.config_schema_snapshot.get_config_snap(field_snap.type_key),
            stack=self.stack.for_field(field_snap_name),
        )

    def for_array(self, index: int) -> "ValidationContext":
        check.int_param(index, "index")
        return ValidationContext(
            config_schema_snapshot=self.config_schema_snapshot,
            config_type_snap=self.config_schema_snapshot.get_config_snap(
                self.config_type_snap.inner_type_key
            ),
            stack=self.stack.for_array_index(index),
        )

    def for_map_key(self, key: object) -> "ValidationContext":
        return ValidationContext(
            config_schema_snapshot=self.config_schema_snapshot,
            config_type_snap=self.config_schema_snapshot.get_config_snap(
                self.config_type_snap.key_type_key
            ),
            stack=self.stack.for_map_key(key),
        )

    def for_map_value(self, key: object) -> "ValidationContext":
        return ValidationContext(
            config_schema_snapshot=self.config_schema_snapshot,
            config_type_snap=self.config_schema_snapshot.get_config_snap(
                self.config_type_snap.inner_type_key
            ),
            stack=self.stack.for_map_value(key),
        )

    def for_new_config_type_key(self, config_type_key: str) -> "ValidationContext":
        check.str_param(config_type_key, "config_type_key")
        return ValidationContext(
            config_schema_snapshot=self.config_schema_snapshot,
            config_type_snap=self.config_schema_snapshot.get_config_snap(config_type_key),
            stack=self.stack,
        )

    def for_nullable_inner_type(self) -> "ValidationContext":
        return ValidationContext(
            config_schema_snapshot=self.config_schema_snapshot,
            config_type_snap=self.config_schema_snapshot.get_config_snap(
                self.config_type_snap.inner_type_key
            ),
            stack=self.stack,
        )


@record
class TraversalContext(ContextData):
    config_type: ConfigType
    do_post_process: bool

    @staticmethod
    def from_config_type(
        config_type: ConfigType,
        stack: EvaluationStackEntry,
        do_post_process: bool,
    ) -> "TraversalContext":
        return TraversalContext(
            config_schema_snapshot=config_type.schema_snapshot,
            config_type_snap=config_type.snapshot,
            config_type=config_type,
            stack=stack,
            do_post_process=do_post_process,
        )

    def for_array(self, index: int) -> "TraversalContext":
        check.int_param(index, "index")
        return TraversalContext(
            config_schema_snapshot=self.config_schema_snapshot,
            config_type_snap=self.config_schema_snapshot.get_config_snap(
                self.config_type_snap.inner_type_key
            ),
            config_type=self.config_type.inner_type,  # type: ignore
            stack=self.stack.for_array_index(index),
            do_post_process=self.do_post_process,
        )

    def for_map(self, key: object) -> "TraversalContext":
        return TraversalContext(
            config_schema_snapshot=self.config_schema_snapshot,
            config_type_snap=self.config_schema_snapshot.get_config_snap(
                self.config_type_snap.inner_type_key
            ),
            config_type=self.config_type.inner_type,  # type: ignore
            stack=self.stack.for_map_value(key),
            do_post_process=self.do_post_process,
        )

    def for_field(self, field_def: Field, field_name: str) -> "TraversalContext":
        return TraversalContext(
            config_schema_snapshot=self.config_schema_snapshot,
            config_type_snap=self.config_schema_snapshot.get_config_snap(field_def.config_type.key),
            config_type=field_def.config_type,
            stack=self.stack.for_field(field_name),
            do_post_process=self.do_post_process,
        )

    def for_nullable_inner_type(self) -> "TraversalContext":
        return TraversalContext(
            config_schema_snapshot=self.config_schema_snapshot,
            config_type_snap=self.config_schema_snapshot.get_config_snap(
                self.config_type_snap.inner_type_key
            ),
            config_type=self.config_type.inner_type,  # type: ignore
            stack=self.stack,
            do_post_process=self.do_post_process,
        )

    def for_new_config_type(self, config_type: ConfigType) -> "TraversalContext":
        return TraversalContext(
            config_schema_snapshot=self.config_schema_snapshot,
            config_type_snap=self.config_schema_snapshot.get_config_snap(config_type.key),
            config_type=config_type,
            stack=self.stack,
            do_post_process=self.do_post_process,
        )



import copy
import logging
import logging.config
import sys
import traceback
import warnings
from collections.abc import Mapping, Sequence
from typing import TYPE_CHECKING, Any, Callable, NamedTuple, Optional, Union

import coloredlogs
import structlog
from typing_extensions import TypeAlias

import dagster._check as check
import dagster._seven as seven
from dagster._annotations import deprecated
from dagster._core.definitions.logger_definition import LoggerDefinition, logger
from dagster._core.utils import coerce_valid_log_level

if TYPE_CHECKING:
    from dagster._core.execution.context.logger import InitLoggerContext


class JsonFileHandler(logging.Handler):
    def __init__(self, json_path: str):
        super().__init__()
        self.json_path = check.str_param(json_path, "json_path")

    def emit(self, record: logging.LogRecord) -> None:
        from dagster._core.log_manager import LOG_RECORD_METADATA_ATTR

        try:
            log_dict = copy.copy(record.__dict__)

            # This horrific monstrosity is to maintain backwards compatability
            # with the old behavior of the JsonFileHandler, which the clarify
            # project has a dependency on. It relied on the dagster-defined
            # properties smashing all the properties of the LogRecord object
            # and uploads all of those properties to a redshift table for
            # in order to do analytics on the log

            if LOG_RECORD_METADATA_ATTR in log_dict:
                dagster_meta_dict = log_dict[LOG_RECORD_METADATA_ATTR]
                del log_dict[LOG_RECORD_METADATA_ATTR]
            else:
                dagster_meta_dict = {}

            log_dict.update(dagster_meta_dict)

            with open(self.json_path, "a", encoding="utf8") as ff:
                text_line = seven.json.dumps(log_dict)
                ff.write(text_line + "\n")
        # Need to catch Exception here, so disabling lint
        except Exception as e:
            logging.critical("[%s] Error during logging!", self.__class__.__name__)
            logging.exception(str(e))


class StructuredLoggerMessage(
    NamedTuple(
        "_StructuredLoggerMessage",
        [
            ("name", str),
            ("message", str),
            ("level", int),
            ("meta", Mapping[str, object]),
            ("record", logging.LogRecord),
        ],
    )
):
    def __new__(
        cls,
        name: str,
        message: str,
        level: int,
        meta: Mapping[str, object],
        record: logging.LogRecord,
    ):
        return super().__new__(
            cls,
            check.str_param(name, "name"),
            check.str_param(message, "message"),
            coerce_valid_log_level(level),
            check.mapping_param(meta, "meta"),
            check.inst_param(record, "record", logging.LogRecord),
        )


StructuredLoggerCallback: TypeAlias = Callable[[StructuredLoggerMessage], None]


class StructuredLoggerHandler(logging.Handler):
    def __init__(self, callback: StructuredLoggerCallback):
        super().__init__()
        self.callback = check.is_callable(callback, "callback")

    def emit(self, record: logging.LogRecord) -> None:
        try:
            self.callback(
                StructuredLoggerMessage(
                    name=record.name,
                    message=record.msg,
                    level=record.levelno,
                    meta=record.dagster_meta,  # type: ignore
                    record=record,
                )
            )
        # Need to catch Exception here, so disabling lint
        except Exception as e:
            logging.critical("[%s] Error during logging!", self.__class__.__name__)
            logging.exception(str(e))


def construct_single_handler_logger(
    name: str, level: Union[str, int], handler: logging.Handler
) -> LoggerDefinition:
    check.str_param(name, "name")
    check.inst_param(handler, "handler", logging.Handler)

    level = coerce_valid_log_level(level)

    @logger
    def single_handler_logger(_init_context: "InitLoggerContext"):
        klass = logging.getLoggerClass()
        logger_ = klass(name, level=level)
        logger_.addHandler(handler)
        handler.setLevel(level)
        return logger_

    return single_handler_logger


# Base python logger whose messages will be captured as structured Dagster log messages.
BASE_DAGSTER_LOGGER = logging.getLogger(name="dagster")


def get_dagster_logger(name: Optional[str] = None) -> logging.Logger:
    """Creates a python logger whose output messages will be captured and converted into Dagster log
    messages. This means they will have structured information such as the step_key, run_id, etc.
    embedded into them, and will show up in the Dagster event log.

    This can be used as a more convenient alternative to `context.log` in most cases. If log level
    is not set explicitly, defaults to DEBUG.

    Args:
        name (Optional[str]): If supplied, will create a logger with the name "dagster.builtin.{name}",
            with properties inherited from the base Dagster logger. If omitted, the returned logger
            will be named "dagster.builtin".

    Returns:
        :class:`logging.Logger`: A logger whose output will be captured by Dagster.

    Example:
        .. code-block:: python

            from dagster import get_dagster_logger, op

            @op
            def hello_op():
                log = get_dagster_logger()
                for i in range(5):
                    # do something
                    log.info(f"Did {i+1} things!")

    """
    # enforce that the parent logger will always have a DEBUG log level
    BASE_DAGSTER_LOGGER.setLevel(logging.DEBUG)
    base_builtin = BASE_DAGSTER_LOGGER.getChild("builtin")
    if name:
        return base_builtin.getChild(name)
    return base_builtin


def define_structured_logger(
    name: str, callback: StructuredLoggerCallback, level: Union[str, int]
) -> LoggerDefinition:
    check.str_param(name, "name")
    check.callable_param(callback, "callback")
    level = coerce_valid_log_level(level)

    return construct_single_handler_logger(name, level, StructuredLoggerHandler(callback))


def define_json_file_logger(name: str, json_path: str, level: Union[str, int]) -> LoggerDefinition:
    check.str_param(name, "name")
    check.str_param(json_path, "json_path")
    level = coerce_valid_log_level(level)

    stream_handler = JsonFileHandler(json_path)
    stream_handler.setFormatter(define_default_formatter())
    return construct_single_handler_logger(name, level, stream_handler)


def get_stack_trace_array(exception: Exception) -> Sequence[str]:
    check.inst_param(exception, "exception", Exception)
    if hasattr(exception, "__traceback__"):
        tb = exception.__traceback__
    else:
        _exc_type, _exc_value, tb = sys.exc_info()
    return traceback.format_tb(tb)


def default_format_string() -> str:
    return "%(asctime)s - %(name)s - %(levelname)s - %(message)s"


def default_date_format_string() -> str:
    return "%Y-%m-%d %H:%M:%S %z"


def define_default_formatter() -> logging.Formatter:
    return logging.Formatter(default_format_string(), default_date_format_string())


def get_structlog_shared_processors():
    timestamper = structlog.processors.TimeStamper(fmt="iso", utc=True)

    shared_processors = [
        structlog.stdlib.add_logger_name,
        structlog.stdlib.add_log_level,
        timestamper,
        structlog.processors.StackInfoRenderer(),
        structlog.stdlib.ExtraAdder(),
    ]

    return shared_processors


def get_structlog_json_formatter() -> structlog.stdlib.ProcessorFormatter:
    return structlog.stdlib.ProcessorFormatter(
        foreign_pre_chain=get_structlog_shared_processors(),
        processors=[
            structlog.stdlib.ProcessorFormatter.remove_processors_meta,
            structlog.processors.JSONRenderer(),
        ],
    )


@deprecated(
    breaking_version="2.0",
    subject="loggers.dagit",
    emit_runtime_warning=False,
)
def configure_loggers(
    handler: str = "default", formatter: str = "colored", log_level: Union[str, int] = "INFO"
) -> None:
    # It's possible that structlog has already been configured by either the user or a controlling
    # process. If so, we don't want to override that configuration.
    if not structlog.is_configured():
        structlog.configure(
            processors=[
                *get_structlog_shared_processors(),
                structlog.stdlib.ProcessorFormatter.wrap_for_formatter,
            ],
            logger_factory=structlog.stdlib.LoggerFactory(),
        )
    json_formatter = get_structlog_json_formatter()

    LOGGING_CONFIG: dict[str, Any] = {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "colored": {
                "()": coloredlogs.ColoredFormatter,
                "fmt": default_format_string(),
                "datefmt": default_date_format_string(),
                "field_styles": {"levelname": {"color": "blue"}, "asctime": {"color": "green"}},
                "level_styles": {"debug": {}, "error": {"color": "red"}},
            },
            "json": {
                "()": json_formatter.__class__,
                "foreign_pre_chain": json_formatter.foreign_pre_chain,
                "processors": json_formatter.processors,
            },
            "rich": {
                "()": structlog.stdlib.ProcessorFormatter,
                "foreign_pre_chain": get_structlog_shared_processors(),
                "processors": [
                    structlog.stdlib.ProcessorFormatter.remove_processors_meta,
                    structlog.dev.ConsoleRenderer(),
                ],
            },
        },
        "handlers": {
            "default": {
                "formatter": formatter,
                "class": "logging.StreamHandler",
                "stream": sys.stdout,
                "level": log_level,
            },
            "null": {
                "class": "logging.NullHandler",
            },
        },
        "loggers": {
            "dagster": {
                "handlers": [handler],
                "level": log_level,
            },
            # Only one of dagster or dagster-webserver will be used at a time. We configure them
            # both here to avoid a dependency on the dagster-webserver package.
            "dagit": {
                "handlers": [handler],
                "level": log_level,
            },
            "dagster-webserver": {
                "handlers": [handler],
                "level": log_level,
            },
        },
    }

    logging.config.dictConfig(LOGGING_CONFIG)

    # override the default warnings handler as per https://docs.python.org/3/library/warnings.html#warnings.showwarning
    # to use the same formatting
    def custom_warning_handler(message, category, filename, lineno, file=None, line=None):
        log_message = warnings.formatwarning(message, category, filename, lineno, line)
        logging.getLogger("dagster").warning(log_message)

    warnings.showwarning = custom_warning_handler


def create_console_logger(name: str, level: Union[str, int]) -> logging.Logger:
    klass = logging.getLoggerClass()
    logger = klass(name, level=level)
    coloredlogs.install(
        logger=logger,
        level=level,
        fmt=default_format_string(),
        datefmt=default_date_format_string(),
        field_styles={"levelname": {"color": "blue"}, "asctime": {"color": "green"}},
        level_styles={"debug": {}, "error": {"color": "red"}},
    )
    return logger



"""add run_id to run_tags index

Revision ID: 6b7fb194ff9c
Revises: 16e3655b4d9b
Create Date: 2024-11-14 22:19:46.456731

"""

from alembic import op
from dagster._core.storage.migration.utils import has_index, has_table

# revision identifiers, used by Alembic.
revision = "6b7fb194ff9c"
down_revision = "16e3655b4d9b"
branch_labels = None
depends_on = None


def upgrade():
    if has_table("run_tags"):
        if not has_index("run_tags", "idx_run_tags_run_id"):
            op.create_index(
                "idx_run_tags_run_id",
                "run_tags",
                ["key", "value", "run_id"],
                unique=True,
                postgresql_concurrently=True,
                mysql_length={"key": 64, "value": 64, "run_id": 255},
            )
        if has_index("run_tags", "idx_run_tags"):
            op.drop_index(
                "idx_run_tags",
                "run_tags",
                postgresql_concurrently=True,
            )


def downgrade():
    if has_table("run_tags"):
        if not has_index("run_tags", "idx_run_tags"):
            op.create_index(
                "idx_run_tags",
                "run_tags",
                ["key", "value"],
                unique=False,
                postgresql_concurrently=True,
                mysql_length={"key": 64, "value": 64},
            )
        if has_index("run_tags", "idx_run_tags_run_id"):
            op.drop_index(
                "idx_run_tags_run_id",
                "run_tags",
                postgresql_concurrently=True,
            )



# @generated

# This file was generated by running `python -m dagster._grpc.compile`
# Do not edit this file directly, and do not attempt to recompile it using
# grpc_tools.protoc directly, as several changes must be made to the raw output

# noqa: SLF001

# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: dagster_api.proto
# Protobuf Python Version: 4.25.1
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x11\x64\x61gster_api.proto\x12\x03\x61pi"\x07\n\x05\x45mpty"\x1b\n\x0bPingRequest\x12\x0c\n\x04\x65\x63ho\x18\x01 \x01(\t"H\n\tPingReply\x12\x0c\n\x04\x65\x63ho\x18\x01 \x01(\t\x12-\n%serialized_server_utilization_metrics\x18\x02 \x01(\t"=\n\x14StreamingPingRequest\x12\x17\n\x0fsequence_length\x18\x01 \x01(\x05\x12\x0c\n\x04\x65\x63ho\x18\x02 \x01(\t";\n\x12StreamingPingEvent\x12\x17\n\x0fsequence_number\x18\x01 \x01(\x05\x12\x0c\n\x04\x65\x63ho\x18\x02 \x01(\t"%\n\x10GetServerIdReply\x12\x11\n\tserver_id\x18\x01 \x01(\t"O\n\x1c\x45xecutionPlanSnapshotRequest\x12/\n\'serialized_execution_plan_snapshot_args\x18\x01 \x01(\t"H\n\x1a\x45xecutionPlanSnapshotReply\x12*\n"serialized_execution_plan_snapshot\x18\x01 \x01(\t"H\n\x1d\x45xternalPartitionNamesRequest\x12\'\n\x1fserialized_partition_names_args\x18\x01 \x01(\t"p\n\x1b\x45xternalPartitionNamesReply\x12Q\nIserialized_external_partition_names_or_external_partition_execution_error\x18\x01 \x01(\t"4\n\x1b\x45xternalNotebookDataRequest\x12\x15\n\rnotebook_path\x18\x01 \x01(\t",\n\x19\x45xternalNotebookDataReply\x12\x0f\n\x07\x63ontent\x18\x01 \x01(\x0c"C\n\x1e\x45xternalPartitionConfigRequest\x12!\n\x19serialized_partition_args\x18\x01 \x01(\t"r\n\x1c\x45xternalPartitionConfigReply\x12R\nJserialized_external_partition_config_or_external_partition_execution_error\x18\x01 \x01(\t"A\n\x1c\x45xternalPartitionTagsRequest\x12!\n\x19serialized_partition_args\x18\x01 \x01(\t"n\n\x1a\x45xternalPartitionTagsReply\x12P\nHserialized_external_partition_tags_or_external_partition_execution_error\x18\x01 \x01(\t"c\n*ExternalPartitionSetExecutionParamsRequest\x12\x35\n-serialized_partition_set_execution_param_args\x18\x01 \x01(\t"\x19\n\x17ListRepositoriesRequest"O\n\x15ListRepositoriesReply\x12\x36\n.serialized_list_repositories_response_or_error\x18\x01 \x01(\t"Y\n%ExternalPipelineSubsetSnapshotRequest\x12\x30\n(serialized_pipeline_subset_snapshot_args\x18\x01 \x01(\t"Y\n#ExternalPipelineSubsetSnapshotReply\x12\x32\n*serialized_external_pipeline_subset_result\x18\x01 \x01(\t"a\n\x19\x45xternalRepositoryRequest\x12+\n#serialized_repository_python_origin\x18\x01 \x01(\t\x12\x17\n\x0f\x64\x65\x66\x65r_snapshots\x18\x02 \x01(\x08"F\n\x17\x45xternalRepositoryReply\x12+\n#serialized_external_repository_data\x18\x01 \x01(\t"i\n StreamingExternalRepositoryEvent\x12\x17\n\x0fsequence_number\x18\x01 \x01(\x05\x12,\n$serialized_external_repository_chunk\x18\x02 \x01(\t"W\n ExternalScheduleExecutionRequest\x12\x33\n+serialized_external_schedule_execution_args\x18\x01 \x01(\t"S\n\x1e\x45xternalSensorExecutionRequest\x12\x31\n)serialized_external_sensor_execution_args\x18\x01 \x01(\t"H\n\x13StreamingChunkEvent\x12\x17\n\x0fsequence_number\x18\x01 \x01(\x05\x12\x18\n\x10serialized_chunk\x18\x02 \x01(\t"@\n\x13ShutdownServerReply\x12)\n!serialized_shutdown_server_result\x18\x01 \x01(\t"E\n\x16\x43\x61ncelExecutionRequest\x12+\n#serialized_cancel_execution_request\x18\x01 \x01(\t"B\n\x14\x43\x61ncelExecutionReply\x12*\n"serialized_cancel_execution_result\x18\x01 \x01(\t"L\n\x19\x43\x61nCancelExecutionRequest\x12/\n\'serialized_can_cancel_execution_request\x18\x01 \x01(\t"I\n\x17\x43\x61nCancelExecutionReply\x12.\n&serialized_can_cancel_execution_result\x18\x01 \x01(\t"6\n\x0fStartRunRequest\x12#\n\x1bserialized_execute_run_args\x18\x01 \x01(\t"4\n\rStartRunReply\x12#\n\x1bserialized_start_run_result\x18\x01 \x01(\t"8\n\x14GetCurrentImageReply\x12 \n\x18serialized_current_image\x18\x01 \x01(\t"6\n\x13GetCurrentRunsReply\x12\x1f\n\x17serialized_current_runs\x18\x01 \x01(\t"L\n\x12\x45xternalJobRequest\x12$\n\x1cserialized_repository_origin\x18\x01 \x01(\t\x12\x10\n\x08job_name\x18\x02 \x01(\t"I\n\x10\x45xternalJobReply\x12\x1b\n\x13serialized_job_data\x18\x01 \x01(\t\x12\x18\n\x10serialized_error\x18\x02 \x01(\t"D\n\x1e\x45xternalScheduleExecutionReply\x12"\n\x1aserialized_schedule_result\x18\x01 \x01(\t"@\n\x1c\x45xternalSensorExecutionReply\x12 \n\x18serialized_sensor_result\x18\x01 \x01(\t"\x13\n\x11ReloadCodeRequest"+\n\x0fReloadCodeReply\x12\x18\n\x10serialized_error\x18\x02 \x01(\t2\xe9\x10\n\nDagsterApi\x12*\n\x04Ping\x12\x10.api.PingRequest\x1a\x0e.api.PingReply"\x00\x12/\n\tHeartbeat\x12\x10.api.PingRequest\x1a\x0e.api.PingReply"\x00\x12G\n\rStreamingPing\x12\x19.api.StreamingPingRequest\x1a\x17.api.StreamingPingEvent"\x00\x30\x01\x12\x32\n\x0bGetServerId\x12\n.api.Empty\x1a\x15.api.GetServerIdReply"\x00\x12]\n\x15\x45xecutionPlanSnapshot\x12!.api.ExecutionPlanSnapshotRequest\x1a\x1f.api.ExecutionPlanSnapshotReply"\x00\x12N\n\x10ListRepositories\x12\x1c.api.ListRepositoriesRequest\x1a\x1a.api.ListRepositoriesReply"\x00\x12`\n\x16\x45xternalPartitionNames\x12".api.ExternalPartitionNamesRequest\x1a .api.ExternalPartitionNamesReply"\x00\x12Z\n\x14\x45xternalNotebookData\x12 .api.ExternalNotebookDataRequest\x1a\x1e.api.ExternalNotebookDataReply"\x00\x12\x63\n\x17\x45xternalPartitionConfig\x12#.api.ExternalPartitionConfigRequest\x1a!.api.ExternalPartitionConfigReply"\x00\x12]\n\x15\x45xternalPartitionTags\x12!.api.ExternalPartitionTagsRequest\x1a\x1f.api.ExternalPartitionTagsReply"\x00\x12t\n#ExternalPartitionSetExecutionParams\x12/.api.ExternalPartitionSetExecutionParamsRequest\x1a\x18.api.StreamingChunkEvent"\x00\x30\x01\x12x\n\x1e\x45xternalPipelineSubsetSnapshot\x12*.api.ExternalPipelineSubsetSnapshotRequest\x1a(.api.ExternalPipelineSubsetSnapshotReply"\x00\x12T\n\x12\x45xternalRepository\x12\x1e.api.ExternalRepositoryRequest\x1a\x1c.api.ExternalRepositoryReply"\x00\x12?\n\x0b\x45xternalJob\x12\x17.api.ExternalJobRequest\x1a\x15.api.ExternalJobReply"\x00\x12h\n\x1bStreamingExternalRepository\x12\x1e.api.ExternalRepositoryRequest\x1a%.api.StreamingExternalRepositoryEvent"\x00\x30\x01\x12`\n\x19\x45xternalScheduleExecution\x12%.api.ExternalScheduleExecutionRequest\x1a\x18.api.StreamingChunkEvent"\x00\x30\x01\x12m\n\x1dSyncExternalScheduleExecution\x12%.api.ExternalScheduleExecutionRequest\x1a#.api.ExternalScheduleExecutionReply"\x00\x12\\\n\x17\x45xternalSensorExecution\x12#.api.ExternalSensorExecutionRequest\x1a\x18.api.StreamingChunkEvent"\x00\x30\x01\x12g\n\x1bSyncExternalSensorExecution\x12#.api.ExternalSensorExecutionRequest\x1a!.api.ExternalSensorExecutionReply"\x00\x12\x38\n\x0eShutdownServer\x12\n.api.Empty\x1a\x18.api.ShutdownServerReply"\x00\x12K\n\x0f\x43\x61ncelExecution\x12\x1b.api.CancelExecutionRequest\x1a\x19.api.CancelExecutionReply"\x00\x12T\n\x12\x43\x61nCancelExecution\x12\x1e.api.CanCancelExecutionRequest\x1a\x1c.api.CanCancelExecutionReply"\x00\x12\x36\n\x08StartRun\x12\x14.api.StartRunRequest\x1a\x12.api.StartRunReply"\x00\x12:\n\x0fGetCurrentImage\x12\n.api.Empty\x1a\x19.api.GetCurrentImageReply"\x00\x12\x38\n\x0eGetCurrentRuns\x12\n.api.Empty\x1a\x18.api.GetCurrentRunsReply"\x00\x12<\n\nReloadCode\x12\x16.api.ReloadCodeRequest\x1a\x14.api.ReloadCodeReply"\x00\x62\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "dagster_api_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    DESCRIPTOR._options = None
    _globals["_EMPTY"]._serialized_start = 26
    _globals["_EMPTY"]._serialized_end = 33
    _globals["_PINGREQUEST"]._serialized_start = 35
    _globals["_PINGREQUEST"]._serialized_end = 62
    _globals["_PINGREPLY"]._serialized_start = 64
    _globals["_PINGREPLY"]._serialized_end = 136
    _globals["_STREAMINGPINGREQUEST"]._serialized_start = 138
    _globals["_STREAMINGPINGREQUEST"]._serialized_end = 199
    _globals["_STREAMINGPINGEVENT"]._serialized_start = 201
    _globals["_STREAMINGPINGEVENT"]._serialized_end = 260
    _globals["_GETSERVERIDREPLY"]._serialized_start = 262
    _globals["_GETSERVERIDREPLY"]._serialized_end = 299
    _globals["_EXECUTIONPLANSNAPSHOTREQUEST"]._serialized_start = 301
    _globals["_EXECUTIONPLANSNAPSHOTREQUEST"]._serialized_end = 380
    _globals["_EXECUTIONPLANSNAPSHOTREPLY"]._serialized_start = 382
    _globals["_EXECUTIONPLANSNAPSHOTREPLY"]._serialized_end = 454
    _globals["_EXTERNALPARTITIONNAMESREQUEST"]._serialized_start = 456
    _globals["_EXTERNALPARTITIONNAMESREQUEST"]._serialized_end = 528
    _globals["_EXTERNALPARTITIONNAMESREPLY"]._serialized_start = 530
    _globals["_EXTERNALPARTITIONNAMESREPLY"]._serialized_end = 642
    _globals["_EXTERNALNOTEBOOKDATAREQUEST"]._serialized_start = 644
    _globals["_EXTERNALNOTEBOOKDATAREQUEST"]._serialized_end = 696
    _globals["_EXTERNALNOTEBOOKDATAREPLY"]._serialized_start = 698
    _globals["_EXTERNALNOTEBOOKDATAREPLY"]._serialized_end = 742
    _globals["_EXTERNALPARTITIONCONFIGREQUEST"]._serialized_start = 744
    _globals["_EXTERNALPARTITIONCONFIGREQUEST"]._serialized_end = 811
    _globals["_EXTERNALPARTITIONCONFIGREPLY"]._serialized_start = 813
    _globals["_EXTERNALPARTITIONCONFIGREPLY"]._serialized_end = 927
    _globals["_EXTERNALPARTITIONTAGSREQUEST"]._serialized_start = 929
    _globals["_EXTERNALPARTITIONTAGSREQUEST"]._serialized_end = 994
    _globals["_EXTERNALPARTITIONTAGSREPLY"]._serialized_start = 996
    _globals["_EXTERNALPARTITIONTAGSREPLY"]._serialized_end = 1106
    _globals["_EXTERNALPARTITIONSETEXECUTIONPARAMSREQUEST"]._serialized_start = 1108
    _globals["_EXTERNALPARTITIONSETEXECUTIONPARAMSREQUEST"]._serialized_end = 1207
    _globals["_LISTREPOSITORIESREQUEST"]._serialized_start = 1209
    _globals["_LISTREPOSITORIESREQUEST"]._serialized_end = 1234
    _globals["_LISTREPOSITORIESREPLY"]._serialized_start = 1236
    _globals["_LISTREPOSITORIESREPLY"]._serialized_end = 1315
    _globals["_EXTERNALPIPELINESUBSETSNAPSHOTREQUEST"]._serialized_start = 1317
    _globals["_EXTERNALPIPELINESUBSETSNAPSHOTREQUEST"]._serialized_end = 1406
    _globals["_EXTERNALPIPELINESUBSETSNAPSHOTREPLY"]._serialized_start = 1408
    _globals["_EXTERNALPIPELINESUBSETSNAPSHOTREPLY"]._serialized_end = 1497
    _globals["_EXTERNALREPOSITORYREQUEST"]._serialized_start = 1499
    _globals["_EXTERNALREPOSITORYREQUEST"]._serialized_end = 1596
    _globals["_EXTERNALREPOSITORYREPLY"]._serialized_start = 1598
    _globals["_EXTERNALREPOSITORYREPLY"]._serialized_end = 1668
    _globals["_STREAMINGEXTERNALREPOSITORYEVENT"]._serialized_start = 1670
    _globals["_STREAMINGEXTERNALREPOSITORYEVENT"]._serialized_end = 1775
    _globals["_EXTERNALSCHEDULEEXECUTIONREQUEST"]._serialized_start = 1777
    _globals["_EXTERNALSCHEDULEEXECUTIONREQUEST"]._serialized_end = 1864
    _globals["_EXTERNALSENSOREXECUTIONREQUEST"]._serialized_start = 1866
    _globals["_EXTERNALSENSOREXECUTIONREQUEST"]._serialized_end = 1949
    _globals["_STREAMINGCHUNKEVENT"]._serialized_start = 1951
    _globals["_STREAMINGCHUNKEVENT"]._serialized_end = 2023
    _globals["_SHUTDOWNSERVERREPLY"]._serialized_start = 2025
    _globals["_SHUTDOWNSERVERREPLY"]._serialized_end = 2089
    _globals["_CANCELEXECUTIONREQUEST"]._serialized_start = 2091
    _globals["_CANCELEXECUTIONREQUEST"]._serialized_end = 2160
    _globals["_CANCELEXECUTIONREPLY"]._serialized_start = 2162
    _globals["_CANCELEXECUTIONREPLY"]._serialized_end = 2228
    _globals["_CANCANCELEXECUTIONREQUEST"]._serialized_start = 2230
    _globals["_CANCANCELEXECUTIONREQUEST"]._serialized_end = 2306
    _globals["_CANCANCELEXECUTIONREPLY"]._serialized_start = 2308
    _globals["_CANCANCELEXECUTIONREPLY"]._serialized_end = 2381
    _globals["_STARTRUNREQUEST"]._serialized_start = 2383
    _globals["_STARTRUNREQUEST"]._serialized_end = 2437
    _globals["_STARTRUNREPLY"]._serialized_start = 2439
    _globals["_STARTRUNREPLY"]._serialized_end = 2491
    _globals["_GETCURRENTIMAGEREPLY"]._serialized_start = 2493
    _globals["_GETCURRENTIMAGEREPLY"]._serialized_end = 2549
    _globals["_GETCURRENTRUNSREPLY"]._serialized_start = 2551
    _globals["_GETCURRENTRUNSREPLY"]._serialized_end = 2605
    _globals["_EXTERNALJOBREQUEST"]._serialized_start = 2607
    _globals["_EXTERNALJOBREQUEST"]._serialized_end = 2683
    _globals["_EXTERNALJOBREPLY"]._serialized_start = 2685
    _globals["_EXTERNALJOBREPLY"]._serialized_end = 2758
    _globals["_EXTERNALSCHEDULEEXECUTIONREPLY"]._serialized_start = 2760
    _globals["_EXTERNALSCHEDULEEXECUTIONREPLY"]._serialized_end = 2828
    _globals["_EXTERNALSENSOREXECUTIONREPLY"]._serialized_start = 2830
    _globals["_EXTERNALSENSOREXECUTIONREPLY"]._serialized_end = 2894
    _globals["_RELOADCODEREQUEST"]._serialized_start = 2896
    _globals["_RELOADCODEREQUEST"]._serialized_end = 2915
    _globals["_RELOADCODEREPLY"]._serialized_start = 2917
    _globals["_RELOADCODEREPLY"]._serialized_end = 2960
    _globals["_DAGSTERAPI"]._serialized_start = 2963
    _globals["_DAGSTERAPI"]._serialized_end = 5116
# @@protoc_insertion_point(module_scope)



import os
from collections import defaultdict
from collections.abc import Mapping, Sequence
from typing import TYPE_CHECKING, Optional

import dagster._check as check
from dagster._core.instance import DagsterInstance
from dagster._core.instance.config import PoolGranularity
from dagster._core.snap.execution_plan_snapshot import ExecutionPlanSnapshot, ExecutionStepSnap
from dagster._core.storage.dagster_run import (
    IN_PROGRESS_RUN_STATUSES,
    DagsterRun,
    DagsterRunStatus,
    RunOpConcurrency,
    RunRecord,
)
from dagster._core.storage.tags import GLOBAL_CONCURRENCY_TAG
from dagster._time import get_current_timestamp

if TYPE_CHECKING:
    from dagster._utils.concurrency import ConcurrencyKeyInfo


def _pool_key_for_step(step: ExecutionStepSnap) -> Optional[str]:
    if step.pool is not None:
        return step.pool

    # for backwards compatibility, we also check the tags
    return (step.tags or {}).get(GLOBAL_CONCURRENCY_TAG)


def compute_run_op_concurrency_info_for_snapshot(
    plan_snapshot: ExecutionPlanSnapshot,
) -> Optional[RunOpConcurrency]:
    """Utility function called at run creation time to add the concurrency info needed to keep track
    of concurrency limits for each in-flight run.
    """
    step_keys_to_execute = set(plan_snapshot.step_keys_to_execute)
    root_step_keys = set(
        [
            step_key
            for step_key, deps in plan_snapshot.step_deps.items()
            if not deps.intersection(step_keys_to_execute)
        ]
    )
    root_key_counts: Mapping[str, int] = defaultdict(int)
    all_pools: set[str] = set()
    has_unconstrained_root_nodes = False
    for step in plan_snapshot.steps:
        if step.key not in step_keys_to_execute:
            continue

        step_pool = _pool_key_for_step(step)
        if step_pool is None and step.key in root_step_keys:
            has_unconstrained_root_nodes = True
        elif step_pool is None:
            continue
        elif step.key in root_step_keys:
            root_key_counts[step_pool] += 1
            if step_pool is not None:
                all_pools.add(step_pool)
        else:
            if step_pool is not None:
                all_pools.add(step_pool)

    if len(all_pools) == 0:
        return None

    return RunOpConcurrency(
        all_pools=all_pools,
        root_key_counts=dict(root_key_counts),
        has_unconstrained_root_nodes=has_unconstrained_root_nodes,
    )


class GlobalOpConcurrencyLimitsCounter:
    def __init__(
        self,
        instance: DagsterInstance,
        runs: Sequence[DagsterRun],
        in_progress_run_records: Sequence[RunRecord],
        slot_count_offset: int = 0,
        pool_granularity: Optional[PoolGranularity] = None,
    ):
        self._root_pools_by_run = {}
        self._concurrency_info_by_key: dict[str, ConcurrencyKeyInfo] = {}
        self._launched_pool_counts = defaultdict(int)
        self._in_progress_pool_counts = defaultdict(int)
        self._slot_count_offset = slot_count_offset
        self._pool_granularity = pool_granularity if pool_granularity else PoolGranularity.OP
        self._in_progress_run_ids: set[str] = set(
            [record.dagster_run.run_id for record in in_progress_run_records]
        )
        self._started_run_pools_allotted_seconds = int(
            os.getenv("DAGSTER_OP_CONCURRENCY_KEYS_ALLOTTED_FOR_STARTED_RUN_SECONDS", "5")
        )

        queued_pool_names = self._get_queued_pool_names(runs)
        # initialize all the pool limits to the default if necessary
        self._initialize_pool_limits(instance, queued_pool_names)

        # fetch all the configured pool keys
        all_configured_pool_names = instance.event_log_storage.get_concurrency_keys()
        configured_queued_pool_names = all_configured_pool_names.intersection(queued_pool_names)

        # fetch all the concurrency info for all of the runs at once, so we can claim in the correct
        # priority order
        self._fetch_concurrency_info(instance, configured_queued_pool_names)

        # fetch all the outstanding pools for in-progress runs
        self._process_in_progress_runs(in_progress_run_records)

    def _get_queued_pool_names(self, queued_runs: Sequence[DagsterRun]) -> set[str]:
        queued_pool_names = set()
        for run in queued_runs:
            if run.run_op_concurrency:
                # if using run granularity, consider all the concurrency keys required by the run
                # if using op granularity, consider only the root keys
                run_pools = (
                    run.run_op_concurrency.root_key_counts.keys()
                    if self._pool_granularity == PoolGranularity.OP
                    else run.run_op_concurrency.all_pools or []
                )
                queued_pool_names.update(run_pools)
        return queued_pool_names

    def _initialize_pool_limits(self, instance: DagsterInstance, pool_names: set[str]):
        default_limit = instance.global_op_concurrency_default_limit
        pool_limits_by_name = {
            pool.name: pool for pool in instance.event_log_storage.get_pool_limits()
        }
        for pool_name in pool_names:
            if pool_name is None:
                continue

            if (pool_name not in pool_limits_by_name and default_limit) or (
                pool_name in pool_limits_by_name
                and pool_limits_by_name[pool_name].from_default
                and pool_limits_by_name[pool_name].limit != default_limit
            ):
                instance.event_log_storage.initialize_concurrency_limit_to_default(pool_name)

    def _fetch_concurrency_info(self, instance: DagsterInstance, pool_names: set[str]):
        for pool_name in pool_names:
            if pool_name is None:
                continue

            self._concurrency_info_by_key[pool_name] = (
                instance.event_log_storage.get_concurrency_info(pool_name)
            )

    def _should_allocate_slots_for_in_progress_run(self, record: RunRecord):
        if not record.dagster_run.run_op_concurrency:
            return False

        status = record.dagster_run.status
        if status not in IN_PROGRESS_RUN_STATUSES:
            return False

        if self._pool_granularity == PoolGranularity.RUN:
            return True

        if status == DagsterRunStatus.STARTING:
            return True

        if status != DagsterRunStatus.STARTED or not record.start_time:
            return False

        time_elapsed = get_current_timestamp() - record.start_time
        if time_elapsed < self._started_run_pools_allotted_seconds:
            return True

        return False

    def _slot_counts_for_run(self, run: DagsterRun) -> Mapping[str, int]:
        if not run.run_op_concurrency:
            return {}

        if self._pool_granularity == PoolGranularity.OP:
            return {**run.run_op_concurrency.root_key_counts}
        elif self._pool_granularity == PoolGranularity.RUN:
            return {pool: 1 for pool in run.run_op_concurrency.all_pools or []}
        else:
            check.failed(f"Unexpected pool granularity {self._pool_granularity}")

    def _process_in_progress_runs(self, in_progress_records: Sequence[RunRecord]):
        for record in in_progress_records:
            if not self._should_allocate_slots_for_in_progress_run(record):
                continue

            for pool, count in self._slot_counts_for_run(record.dagster_run).items():
                self._in_progress_pool_counts[pool] += count

    def is_blocked(self, run: DagsterRun) -> bool:
        # if any of the ops in the run can make progress (not blocked by concurrency keys), we
        # should dequeue
        if not run.run_op_concurrency:
            return False

        if (
            self._pool_granularity == PoolGranularity.OP
            and run.run_op_concurrency.has_unconstrained_root_nodes
        ):
            # if the granularity is at the op level and there exists a root node that is not
            # concurrency blocked, we should dequeue.
            return False

        if self._pool_granularity == PoolGranularity.OP:
            # we just need to check all of the root concurrency keys, instead of all the concurrency keys
            # in the run
            for pool in run.run_op_concurrency.root_key_counts.keys():
                if pool not in self._concurrency_info_by_key:
                    # there is no concurrency limit set for this key, we should dequeue
                    return False

                key_info = self._concurrency_info_by_key[pool]
                available_count = (
                    key_info.slot_count
                    - len(key_info.pending_steps)
                    - self._launched_pool_counts[pool]
                    - self._in_progress_pool_counts[pool]
                )
                if available_count + self._slot_count_offset > 0:
                    # there exists a root concurrency key that is not blocked, we should dequeue
                    return False

            # if we reached here, then every root concurrency key is blocked, so we should not dequeue
            return True

        elif self._pool_granularity == PoolGranularity.RUN:
            # if the granularity is at the run level, we should check if any of the concurrency
            # keys are blocked
            for pool in run.run_op_concurrency.all_pools or []:
                if pool not in self._concurrency_info_by_key:
                    # there is no concurrency limit set for this key
                    continue

                key_info = self._concurrency_info_by_key[pool]
                available_count = (
                    key_info.slot_count
                    - self._launched_pool_counts[pool]
                    - self._in_progress_pool_counts[pool]
                )
                if available_count + self._slot_count_offset <= 0:
                    return True

            # if we reached here then there is at least one available slot for every single concurrency key
            # required by this run, so we should dequeue
            return False
        else:
            check.failed(f"Unexpected pool granularity {self._pool_granularity}")

    def get_blocked_run_debug_info(self, run: DagsterRun) -> Mapping:
        if not run.run_op_concurrency:
            return {}

        log_info = {}
        for pool in run.run_op_concurrency.root_key_counts.keys():
            concurrency_info = self._concurrency_info_by_key.get(pool)
            if not concurrency_info:
                continue

            log_info[pool] = {
                "granularity": self._pool_granularity.value,
                "slot_count": concurrency_info.slot_count,
                "pending_step_count": len(concurrency_info.pending_steps),
                "pending_step_run_ids": list(
                    {step.run_id for step in concurrency_info.pending_steps}
                ),
                "launched_count": self._launched_pool_counts[pool],
                "in_progress_count": self._in_progress_pool_counts[pool],
            }
        return log_info

    def update_counters_with_launched_item(self, run: DagsterRun):
        for pool, count in self._slot_counts_for_run(run).items():
            self._launched_pool_counts[pool] += count
